import datetime
import multiprocessing

import numpy as np
import pandas as pd
from pathos.multiprocessing import ProcessPool


def dataframe_mp(dataframe, func, mp_cores: int = None):
    """
    do multiprocessing on a dataframe

    :param dataframe: input dataframe
    :param func: function to apply to the dataframe
    :param mp_cores: number of cpus to use
    :return: output dataframe
    """

    if mp_cores is None:
        mp_cores = multiprocessing.cpu_count() - 1

    pool = ProcessPool(mp_cores)

    split_input_dataset = np.array_split(dataframe, mp_cores)

    output_dataset = pool.map(func, split_input_dataset)
    del [split_input_dataset]

    output_dataset = pd.concat(output_dataset, ignore_index=True)

    return output_dataset


def get_sunday(date, input_format='%Y%m%d', output_format='%Y%m%d'):
    """
    get the sunday of the week of the input date

    :param date: input date
    :param input_format: input date format, default is yyyymmdd
    :param output_format: output date format (sunday), default is yyyymmdd
    :return: sunday of the week of the input date
    """

    duty_date = datetime.datetime.strptime(str(date), input_format)
    sunday = duty_date

    one_day = datetime.timedelta(days=1)

    while sunday.weekday() != 6:
        sunday += one_day

    return datetime.datetime.strftime(sunday, output_format)


def lambda_groupby(lambda_func, groupby_df, groupby_cols, result_col, reformat=False):
    """
    apply lambda function to a groupby dataframe

    :param lambda_func: lambda function, e.g. lambda x: (x['daily_price'] * x['daily_unit']).sum() / x['daily_unit'].sum()
    :param groupby_df: input groupby dataframe
    :param groupby_cols: columns to groupby
    :param result_col: result column name
    :return: output dataframe
    """

    df_temp = groupby_df.groupby(groupby_cols).apply(lambda_func)
    df_temp = df_temp.to_dict()

    temp_lst = []

    for key, value in df_temp.items():
        a = key[0]
        b = key[1]
        c = value
        temp_lst.append([a, b, c])

    new_col_lst = groupby_cols + [result_col]
    df_temp = pd.DataFrame(data=np.array(temp_lst), columns=new_col_lst)

    if reformat == 'float':
        df_temp[result_col] = df_temp[result_col].astype(float)
    elif reformat == 'int':
        df_temp[result_col] = df_temp[result_col].astype(int)
    else:
        pass

    return df_temp


def calculate_r2_value(input_df, true_col, predict_col, mean_col):
    """
    calculate R2 value

    :param input_df: input dataframe
    :param true_col: the column name of true value
    :param predict_col: the column name of predicted value
    :param mean_col: the column name of mean value
    :return: R2 value
    """

    check = input_df.copy()

    check['a'] = check[true_col] - check[predict_col]
    check['a'] = check['a'].map(lambda x: x * x)
    check['b'] = check[true_col] - check[mean_col]
    check['b'] = check['b'].map(lambda x: x * x)

    R2_VALUE = 1 - check['a'].sum() / check['b'].sum()

    return R2_VALUE



from karma.funcs import calculate_r2_value, dataframe_mp, get_sunday, lambda_groupby
from karma.read_data import read_big_data, read_small_data, save_data
from karma.reformat_data import col_to_float, col_to_str, yaml_to_object



from __future__ import annotations

import os
import time
from pathlib import Path

import pandas as pd
import pyarrow.parquet as pq
from tqdm.auto import tqdm

from .log import logger


def my_read_parquet(file_path, **kwargs) -> pd.DataFrame:
    """
    customized function to read parquet

    Args:
        file_path: parquet file path
        read_cols: columns to read
    Returns:
        loaded pd.dataframe
    """

    parquet_file = pq.ParquetFile(file_path, **kwargs)
    dfs = [batch.to_pandas() for batch in tqdm(parquet_file.iter_batches(), disable=False, desc='Loading data')]
    df = pd.concat(dfs, ignore_index=True)

    return df


def read_small_data(file_name: str, data_folder: str | Path, reformat_cols: str = None, rename_cols: dict = None, **kwargs) -> pd.DataFrame:
    """
    customized function to read a singe data file, supporting read_excel / read_csv / read_parquet

    Args:
        file_name: full file path including suffix
        data_folder: the main dir where the data is saved
        sub_folder: the sub folder name where the data is saved
        csv_chunksize: chunk size for read_csv, default is None
        reformat_cols: reformat column names to lower or upper characters, accept `lower` / `upper`, default is None
        read_cols: columns to read
        rename_cols: rename the columns
        special_sep: special sep for .csv file
    Returns:
        loaded pd.dataframe
    """

    data_folder = Path(data_folder)
    file_path = data_folder.joinpath(file_name)

    logger.info(f'Reading {file_path}')
    start_time = time.time()

    file_suffix = file_name.split('.', 1)[1]

    if file_suffix in ['xlsx', 'xls']:
        df = pd.read_excel(file_path, **kwargs)

    elif file_suffix == 'csv':
        if 'nrows' in kwargs:
            df = pd.read_csv(file_path, low_memory=False, **kwargs)
        else:
            rows = sum(1 for _ in open(file_path, 'r')) - 1

            chunk_list = []
            with tqdm(total=rows, desc='Loading data') as bar:
                for chunk in pd.read_csv(file_path, chunksize=10000, low_memory=False, **kwargs):
                    chunk_list.append(chunk)
                    bar.update(len(chunk))

            df = pd.concat((f for f in chunk_list), axis=0)

    elif file_suffix == 'txt':
        df = df = pd.read_csv(file_path, low_memory=False, **kwargs)

    elif file_suffix == 'gzip':
        df = my_read_parquet(file_path=file_path, **kwargs)

    elif 'csv.' in file_suffix:
        compression_type = file_suffix.split('.', 1)[1]

        if '.' in compression_type:
            raise ValueError(f'invalid file name {file_name}')
        else:
            if compression_type == 'gz':
                compression_type = 'gzip'
            else:
                pass

            df = pd.read_csv(file_path, low_memory=False, compression=compression_type, **kwargs)

    else:
        raise ValueError(f'invalid file name {file_name}')

    if rename_cols is None:
        pass
    else:
        df.rename(columns=rename_cols, inplace=True)

    if reformat_cols == 'upper':
        df.rename(columns=str.upper, inplace=True)
    elif reformat_cols == 'lower':
        df.rename(columns=str.lower, inplace=True)
    else:
        pass

    logger.info(
        f'Data Reading Time: {round((time.time() - start_time), 2)} seconds, ' f'read {df.shape[0]} rows and columns are: {df.columns.to_list()}'
    )

    return df


def read_big_data(file_name: str, data_folder: str | Path, reformat_cols: str = None, rename_cols: dict = None, **kwargs) -> pd.DataFrame:
    """
    customized function to read a singe data file, supporting read_excel / read_csv / read_parquet

    Args:
        file_name: full file path including suffix
        data_folder: the main dir where the data is saved
        reformat_cols: reformat column names to lower or upper characters, accept `lower` / `upper`, default is None
        rename_cols: rename the columns
    Returns:
        loaded pd.dataframe
    """

    data_folder = Path(data_folder)

    file_path = data_folder.joinpath(file_name)

    logger.info(f'Reading {file_path}')
    start_time = time.time()

    file_suffix = file_name.split('.', 1)[1]

    if file_suffix in ['xlsx', 'xls']:
        df = pd.read_excel(file_path, **kwargs)

    elif file_suffix == 'csv':
        if 'nrows' in kwargs:
            df = pd.read_csv(file_path, low_memory=False, **kwargs)
        else:
            logger.info('Loading Chunks ...')
            chunk_reader = pd.read_csv(file_path, chunksize=500000, **kwargs)  # the number of rows per chunk

            df_lst = []
            for df in chunk_reader:
                df_lst.append(df)

            logger.info('Concat Chunks ...')
            df = pd.concat(df_lst, sort=False)

    elif file_suffix == 'txt':
        df = df = pd.read_csv(file_path, low_memory=False, **kwargs)

    elif file_suffix == 'gzip':
        df = my_read_parquet(file_path=file_path, **kwargs)

    elif 'csv.' in file_suffix:
        compression_type = file_suffix.split('.', 1)[1]

        if '.' in compression_type:
            raise ValueError(f'invalid file name {file_name}')
        else:
            if compression_type == 'gz':
                compression_type = 'gzip'
            else:
                pass

            df = pd.read_csv(file_path, low_memory=False, compression=compression_type, **kwargs)

    else:
        raise ValueError(f'invalid file name {file_name}')

    if rename_cols is None:
        pass
    else:
        df.rename(columns=rename_cols, inplace=True)

    if reformat_cols == 'upper':
        df.rename(columns=str.upper, inplace=True)
    elif reformat_cols == 'lower':
        df.rename(columns=str.lower, inplace=True)
    else:
        pass

    logger.info(
        f'Data Reading Time: {round((time.time() - start_time), 2)} seconds, ' f'read {df.shape[0]} rows and columns are: {df.columns.to_list()}'
    )

    return df


def save_data(df: pd.DataFrame, data_folder: str | Path, file_name: str, **kwargs) -> None:
    """
    customized function to save a dataframe, supporting to_excel / to_csv / to_parquet

    Args:
        df: pd.dataframe to be saved
        data_folder: the main dir where the data is saved
        file_name: full file path including suffix
    Returns:
        loaded pd.dataframe
    """

    data_folder = Path(data_folder)

    if data_folder.is_dir():
        pass
    else:
        os.makedirs(data_folder)

    file_path = data_folder.joinpath(file_name)
    file_suffix = file_name.split('.', 1)[1]

    logger.info(f'Saving {file_path}')
    start_time = time.time()

    if 'index' in list(kwargs.keys()):
        pass
    else:
        kwargs['index'] = False

    if file_suffix in ['xlsx', 'xls']:
        df.to_excel(file_path, **kwargs)

    elif file_suffix == 'csv':
        df.to_csv(file_path, **kwargs)

    elif file_suffix == 'gzip':
        df.to_parquet(file_path, compression='gzip', **kwargs)

    elif 'csv.' in file_suffix:
        compression_type = file_suffix.split('.', 1)[1]

        if '.' in compression_type:
            raise ValueError(f'invalid file name {file_name}')
        else:
            df.to_csv(path_or_buf=file_path, compression=compression_type, **kwargs)

    else:
        raise ValueError(f'invalid file name {file_name}')

    logger.info(
        f'Time consumed: {round((time.time() - start_time), 2)} seconds, ' f'saved {df.shape[0]} rows and columns are: {df.columns.to_list()}'
    )



import logging
import sys

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

__formatter = logging.Formatter(fmt='%(asctime)s [%(levelname)s] [%(filename)s:%(lineno)d] - %(message)s')

__console = logging.StreamHandler(stream=sys.stdout)
__console.setLevel(logging.INFO)
__console.setFormatter(__formatter)
logger.addHandler(__console)



from __future__ import annotations

from collections import namedtuple
from pathlib import Path

import numpy as np

from .log import logger


def yaml_to_object(yaml_file, yaml_folder=None, to_object=True) -> namedtuple:
    """
    read yaml config file to a dict

    Args:
        yaml_file: yaml file name
        yaml_folder: yaml file folder
        to_object: whether to transform it to a Python object
    Returns:
        a namedtuple
    """

    if yaml_folder is None:
        yaml_folder = Path(__file__).parent.joinpath('config')
    else:
        yaml_folder = Path(yaml_folder)
    import yaml

    with open(yaml_folder.joinpath(yaml_file), encoding='utf-8') as file:
        data = yaml.safe_load(file)

    if to_object:
        data = namedtuple('ObjectName', data.keys())(*data.values())

    return data


def col_to_str(df, col_list):
    """
    transform int / float column to str

    Args:
        df: dataframe to be transformed
        col_list: list of columns to be transformed
    Returns:
        transformed dataframe
    """

    for col in col_list:
        if df[col].dtype is np.dtype('float') or df[col].dtype is np.dtype('float64'):
            logger.info(f'column {col} is float, changing to str ...')
            df = df.astype({col: np.int64})
            df = df.astype({col: str})
        elif df[col].dtype is np.dtype('int') or df[col].dtype is np.dtype('int64'):
            logger.info(f'column {col} is int, changing to str ...')
            df = df.astype({col: str})
        elif df[col].dtype is np.dtype('O'):
            logger.info(f'column {col} is already str')
        else:
            raise ValueError(f'Not able to handle {df[col].dtype} type')

    return df


def col_to_float(df, col_list):
    """
    transform int / str column to float

    Args:
        df: dataframe to be transformed
        col_list: list of columns to be transformed
    Returns:
        transformed dataframe
    """

    for col in col_list:
        if df[col].dtype is np.dtype('float') or df[col].dtype is np.dtype('float64'):
            logger.info(f'column {col} is already float')
        elif df[col].dtype is np.dtype('int') or df[col].dtype is np.dtype('int64'):
            logger.info(f'column {col} is int, changing to str ...')
            df = df.astype({col: float})
        elif df[col].dtype is np.dtype('O'):
            logger.info(f'column {col} is str, changing to float ...')
            df = df.astype({col: float})
        else:
            raise ValueError(f'Not able to handle {df[col].dtype} type')

    return df
